<?php
namespace Blocks;

/**
 * Asset source base class
 */
abstract class BaseAssetSourceType extends BaseSavableComponentType
{
	/**
	 * @access protected
	 * @var string The type of component this is
	 */
	protected $componentType = 'AssetSourceType';

	/**
	 * Starts an indexing session
	 * @param $sessionId
	 * @return array
	 */
	abstract public function startIndex($sessionId);

	/**
	 * Process an indexing session
	 * @param $sessionId
	 * @param $offset
	 * @return mixed
	 */
	abstract public function processIndex($sessionId, $offset);

	/**
	 * Get the image source path with the optional handle name.
	 *
	 * @param AssetFileModel $fileModel
	 * @return mixed
	 */
	public abstract function getImageSourcePath(AssetFileModel $fileModel);

	/**
	 * Get the timestamp of when a file transformation was last modified.
	 *
	 * @param AssetFileModel $fileModel
	 * @param string $transformationHandle
	 * @return mixed
	 */
	public abstract function getTimeTransformationModified(AssetFileModel $fileModel, $transformationHandle);

	/**
	 * Insert a file from path in folder.
	 *
	 * @param AssetFolderModel $folder
	 * @param $filePath
	 * @param $fileName
	 * @return AssetOperationResponseModel
	 * @throws Exception
	 */
	abstract protected function _insertFileInFolder(AssetFolderModel $folder, $filePath, $fileName);

	/**
	 * Get a name replacement for a filename already taken in a folder.
	 *
	 * @param AssetFolderModel $folder
	 * @param $fileName
	 * @return mixed
	 */
	abstract protected function _getNameReplacement(AssetFolderModel $folder, $fileName);

	/**
	 * Put an image transformation for the File and handle using the provided path to the source image.
	 *
	 * @param AssetFileModel $fileModel
	 * @param $handle
	 * @param $sourceImage
	 * @return mixed
	 */
	abstract public function putImageTransformation(AssetFileModel $fileModel, $handle, $sourceImage);

	/**
	 * Make a local copy of the file and return the path to it.
	 *
	 * @param AssetFileModel $file
	 * @return mixed
	 */
	abstract protected function _getLocalCopy(AssetFileModel $file);

	/**
	 * Delete just the file inside of a source for an Assets File.
	 *
	 * @param AssetFileModel $file
	 */
	abstract protected function _deleteSourceFile(AssetFileModel $file);

	/**
	 * Move a file in source.
	 *
	 * @param AssetFileModel $file
	 * @param AssetFolderModel $targetFolder
	 * @param string $fileName
	 * @param string $userResponse Conflict resolution response
	 * @return mixed
	 */
	abstract protected function _moveSourceFile(AssetFileModel $file, AssetFolderModel $targetFolder, $fileName = '', $userResponse = '');

	/**
	 * Delete generated image transformations for a File.
	 *
	 * @param AssetFileModel $file
	 * @return mixed
	 */
	abstract protected function _deleteGeneratedImageTransformations(AssetFileModel $file);

	/**
	 * Return TRUE if a physical folder exists.
	 *
	 * @param AssetFolderModel $parentFolder
	 * @param $folderName
	 * @return boolean
	 */
	abstract protected function _sourceFolderExists(AssetFolderModel $parentFolder, $folderName);

	/**
	 * Create a physical folder, return TRUE on success.
	 *
	 * @param AssetFolderModel $parentFolder
	 * @param $folderName
	 * @return boolean
	 */
	abstract protected function _createSourceFolder(AssetFolderModel $parentFolder, $folderName);

	/**
	 * Delete the source folder.
	 *
	 * @param AssetFolderModel $folder
	 * @return boolean
	 */
	abstract protected function _deleteSourceFolder(AssetFolderModel $folder);

	/**
	 * Rename a source folder.
	 *
	 * @param AssetFolderModel $folder
	 * @param $newName
	 * @return boolean
	 */
	abstract protected function _renameSourceFolder(AssetFolderModel $folder, $newName);


	/**
	 * Return a result object for prompting the user about filename conflicts.
	 *
	 * @param string $fileName the cause of all trouble
	 * @return object
	 */
	protected function _getUserPromptOptions($fileName)
	{
		return (object) array(
			'message' => Blocks::t('File "{file}" already exists at target location', array('file' => $fileName)),
			'choices' => array(
				array('value' => AssetsHelper::ActionKeepBoth, 'title' => Blocks::t('Rename the new file and keep both')),
				array('value' => AssetsHelper::ActionReplace, 'title' => Blocks::t('Replace the existing file')),
				array('value' => AssetsHelper::ActionCancel, 'title' => Blocks::t('Keep the original file'))
			)
		);
	}

	/**
	 * Returns true if this is a valid source. Used for type-specific validations
	 * @return array
	 */
	public function getSourceErrors()
	{
		return array();
	}

	/**
	 * Upload a file.
	 *
	 * @param AssetFolderModel $folder
	 * @return object
	 * @throws Exception
	 */
	public function uploadFile($folder)
	{
		// Upload the file and drop it in the temporary folder
		$uploader = new \qqFileUploader();

		// Make sure a file was uploaded
		if (! $uploader->file)
		{
			throw new Exception(Blocks::t('No file was uploaded'));
		}

		$size = $uploader->file->getSize();

		// Make sure the file isn't empty
		if (!$size)
		{
			throw new Exception(Blocks::t('Uploaded file was empty'));
		}

		// Save the file to a temp location and pass this on to the source type implementation
		$filePath = AssetsHelper::getTempFilePath(IOHelper::getExtension($uploader->file->getName()));
		$uploader->file->save($filePath);


		$response = $this->_insertFileInFolder($folder, $filePath, $uploader->file->getName());

		// Naming conflict. create a new file and ask the user what to do with it
		if ($response->isConflict())
		{
			$newFileName = $this->_getNameReplacement($folder, $uploader->file->getName());
			$conflictResponse = $response;
			$response = $this->_insertFileInFolder($folder, $filePath, $newFileName);
		}

		if ($response->isSuccess())
		{
			$filename = pathinfo($response->getDataItem('filePath'), PATHINFO_BASENAME);

			$fileModel = new AssetFileModel();
			$fileModel->sourceId = $this->model->id;
			$fileModel->folderId = $folder->id;
			$fileModel->filename = pathinfo($filename, PATHINFO_BASENAME);
			$fileModel->kind = IOHelper::getFileKind(pathinfo($filename, PATHINFO_EXTENSION));
			$fileModel->size = filesize($filePath);
			$fileModel->dateModified = IOHelper::getLastTimeModified($filePath);

			if ($fileModel->kind == 'image')
			{
				list ($width, $height) = getimagesize($filePath);
				$fileModel->width = $width;
				$fileModel->height = $height;
			}

			blx()->assets->storeFile($fileModel);

			if ($this->model->type != 'Local')
			{
				// Store copy locally for all sorts of operations.
				IOHelper::copyFile($filePath, blx()->path->getAssetsImageSourcePath().$fileModel->id.'.'.pathinfo($fileModel, PATHINFO_EXTENSION));
			}

			blx()->assetTransformations->updateTransformations($fileModel, array_keys(blx()->assetTransformations->getAssetTransformations()));

			// Check if we stored a conflict response originally - send that back then.
			if (isset($conflictResponse))
			{
				$response = $conflictResponse;
				$response->setDataItem('additionalInfo', $folder->id.':'.$fileModel->id);
				$response->setDataItem('newFileId', $fileModel->id);
			}

			$response->setDataItem('fileId', $fileModel->id);
		}
		else
		{
			IOHelper::deleteFile($filePath);
		}

		// Prevent sensitive information leak. Just in case.
		$response->deleteDataItem('filePath');

		return $response;
	}

	/**
	 * Ensure a folder entry exists in the DB for the full path and return it's id.
	 *
	 * @param $fullPath
	 * @return int
	 */
	protected function _ensureFolderByFulPath($fullPath)
	{
		$parameters = new FolderCriteriaModel(array(
			'fullPath' => $fullPath,
			'sourceId' => $this->model->id
		));

		$folderModel = blx()->assets->findFolder($parameters);

		// If we don't have a folder matching these, create a new one
		if (is_null($folderModel))
		{
			$parts = explode('/', rtrim($fullPath, '/'));
			$folderName = array_pop($parts);

			if (empty($parts))
			{
				// Looking for a top level folder, apparently.
				$parameters->fullPath = "";
				$parameters->parentId = FolderCriteriaModel::AssetsNoParent;
			}
			else
			{
				$parameters->fullPath = join('/', $parts) . '/';
			}

			// Look up the parent folder
			$parentFolder = blx()->assets->findFolder($parameters);
			if (is_null($parentFolder))
			{
				$parentId = FolderCriteriaModel::AssetsNoParent;
			}
			else
			{
				$parentId = $parentFolder->id;
			}

			$folderModel = new AssetFolderModel();
			$folderModel->sourceId = $this->model->id;
			$folderModel->parentId = $parentId;
			$folderModel->name = $folderName;
			$folderModel->fullPath = $fullPath;

			return blx()->assets->storeFolder($folderModel);
		}
		else
		{
			return $folderModel->id;
		}
	}

	/**
	 * Return a list of missing folders, when comparing the full folder list for this source against the provided list.
	 *
	 * @param array $folderList
	 * @return array
	 */
	protected function _getMissingFolders(array $folderList)
	{
		// Figure out the obsolete records for folders
		$missingFolders = array();

		$allFolders = blx()->assets->findFolders(array(
			'sourceId' => $this->model->id
		));

		foreach ($allFolders as $folderModel)
		{
			if (!isset($folderList[$folderModel->id]))
			{
				$missingFolders[$folderModel->id] = $this->model->name . '/' . $folderModel->fullPath;
			}
		}

		return $missingFolders;
	}

	/**
	 * @param $uriPath
	 * @return AssetFileModel|bool|null
	 */
	protected function _indexFile($uriPath)
	{
		$extension = IOHelper::getExtension($uriPath);

		if (IOHelper::isExtensionAllowed($extension))
		{
			$parts = explode('/', $uriPath);
			$fileName = array_pop($parts);

			$searchFullPath = join('/', $parts) . (empty($parts) ? '' : '/');

			if (empty($searchFullPath))
			{
				$parentId = FolderCriteriaModel::AssetsNoParent;
			}
			else
			{
				$parentId = false;
			}

			$parentFolder = blx()->assets->findFolder(array(
				'sourceId' => $this->model->id,
				'fullPath' => $searchFullPath,
				'parentId' => $parentId
			));

			if (empty($parentFolder))
			{
				return false;
			}

			$folderId = $parentFolder->id;

			$fileModel = blx()->assets->findFile(array(
				'folderId' => $folderId,
				'filename' => $fileName
			));

			if (is_null($fileModel))
			{
				$fileModel = new AssetFileModel();
				$fileModel->sourceId = $this->model->id;
				$fileModel->folderId = $folderId;
				$fileModel->filename = $fileName;
				$fileModel->kind = IOHelper::getFileKind($extension);
				blx()->assets->storeFile($fileModel);
			}

			return $fileModel;
		}

		return false;
	}

	/**
	 * Replace physical file.
	 *
	 * @param AssetFileModel $oldFile
	 * @param AssetFileModel $replaceWith
	 */
	public function replaceFile(AssetFileModel $oldFile, AssetFileModel $replaceWith)
	{
		if ($oldFile->kind == 'image')
		{
			// we'll need this if replacing images
			$localCopy = $this->_getLocalCopy($replaceWith);
			$this->_deleteGeneratedThumbnails($oldFile);
			$this->_deleteGeneratedImageTransformations($oldFile);
		}

		$this->_deleteSourceFile($oldFile);

		$this->_moveSourceFile($replaceWith, blx()->assets->getFolderById($oldFile->folderId), $oldFile->filename);

		$oldFile->width = $replaceWith->width;
		$oldFile->height = $replaceWith->height;
		$oldFile->size = $replaceWith->size;
		$oldFile->dateModified = $replaceWith->dateModified;

		blx()->assets->storeFile($oldFile);
	}

	/**
	 * Delete all the generated images for this file.
	 *
	 * @param AssetFileModel $file
	 */
	protected function _deleteGeneratedThumbnails(AssetFileModel $file)
	{
		$thumbFolders = IOHelper::getFolderContents(blx()->path->getAssetsThumbsPath());
		foreach ($thumbFolders as $folder)
		{
			if (is_dir($folder))
			{
				IOHelper::deleteFile($folder.'/'.$file->id.'.'.IOHelper::getExtension($file->filename));
			}
		}
	}

	/**
	 * Delete a file.
	 *
	 * @param AssetFileModel $file
	 * @return AssetOperationResponseModel
	 */
	public function deleteFile(AssetFileModel $file)
	{
		$this->_deleteSourceFile($file);
		$this->_deleteGeneratedImageTransformations($file);
		$this->_deleteGeneratedThumbnails($file);

		blx()->assets->deleteFileRecord($file->id);

		$response = new AssetOperationResponseModel();
		$response->setSuccess();

		return $response;
	}

	/**
	 * Create a subfolder.
	 *
	 * @param AssetFolderModel $parentFolder
	 * @param $folderName
	 * @return AssetOperationResponseModel
	 * @throws Exception
	 */
	public function createFolder(AssetFolderModel $parentFolder, $folderName)
	{
		$folderName = IOHelper::cleanFilename($folderName);

		// If folder exists in DB or physically, bail out
		if (blx()->assets->findFolder(array('parentId' => $parentFolder->id, 'name' => $folderName))
			|| $this->_sourceFolderExists($parentFolder, $folderName))
		{
			throw new Exception(Blocks::t('A folder already exists with that name!'));
		}

		if ( !$this->_createSourceFolder($parentFolder, $folderName))
		{
			throw new Exception(Blocks::t('There was an error while creating the folder.'));
		}

		$newFolder = new AssetFolderModel();
		$newFolder->sourceId = $parentFolder->sourceId;
		$newFolder->parentId = $parentFolder->id;
		$newFolder->name = $folderName;
		$newFolder->fullPath = $parentFolder->fullPath.$folderName.'/';

		$folderId = blx()->assets->storeFolder($newFolder);

		$response = new AssetOperationResponseModel();
		$response->setSuccess();
		$response->setDataItem('folderId', $folderId);
		$response->setDataItem('parentId', $parentFolder->id);
		$response->setDataItem('folderName', $folderName);

		return $response;
	}

	/**
	 * Rename a folder.
	 *
	 * @param AssetFolderModel $folder
	 * @param $newName
	 * @return AssetOperationResponseModel
	 */
	public function renameFolder(AssetFolderModel $folder, $newName)
	{
		$parentFolder = blx()->assets->getFolderById($folder->parentId);
		if (!$parentFolder)
		{
			throw new Exception(Blocks::t("Cannot rename folder \"{folder}\"!", array('folder' => $folder->name)));
		}
		if ($this->_sourceFolderExists($parentFolder, $newName))
		{
			throw new Exception(Blocks::t("Folder \"{folder}\" already exists there.", array('folder' => $newName)));
		}

		// Try to rename the folder in the source
		if (!$this->_renameSourceFolder($folder, $newName))
		{
			throw new Exception(Blocks::t("Cannot rename folder \"{folder}\"!", array('folder' => $folder->name)));
		}

		$oldFullPath = $folder->fullPath;
		$newFullPath = $this->_getParentFullPath($folder->fullPath).$newName.'/';

		// Find all folders with affected fullPaths and update them.
		$folders = blx()->assets->findChildFolders($folder);
		foreach ($folders as $folderModel)
		{
			$folderModel->fullPath = preg_replace('#^'.$oldFullPath.'#', $newFullPath, $folderModel->fullPath);
			blx()->assets->storeFolder($folderModel);
		}

		// Now change the affected folder
		$folder->name = $newName;
		$folder->fullPath = $newFullPath;
		blx()->assets->storeFolder($folder);

		// All set, Scotty!
		$response = new AssetOperationResponseModel();
		$response->setSuccess();
		$response->setDataItem('newName', $newName);
		return $response;
	}

	/**
	 * Delete a folder.
	 *
	 * @param AssetFolderModel $folder
	 * @return AssetOperationResponseModel
	 */
	public function deleteFolder(AssetFolderModel $folder)
	{

		// Get rid of children files
		$files = blx()->assets->findFiles(array(
			'folderId' => $folder->id
		));

		foreach ($files as $file)
		{
			$this->deleteFile($file);
		}

		// Delete children folders
		$childFolders = blx()->assets->findFolders(array('parentId' => $folder->id));
		foreach ($childFolders as $childFolder)
		{
			$this->deleteFolder($childFolder);
		}

		$this->_deleteSourceFolder($folder);

		blx()->assets->deleteFolderRecord($folder->id);

		$response = new AssetOperationResponseModel();
		$response->setSuccess();
		return $response;
	}

	/**
	 * Return a parent folder's full path for a full path.
	 *
	 * @param $fullPath
	 * @return string
	 */
	protected function _getParentFullPath($fullPath)
	{
		// Drop the trailing slash and split it by slash
		$parts = explode("/", rtrim($fullPath, "/"));

		// Drop the last part and return the part leading up to it
		array_pop($parts);

		if (empty($parts))
		{
			return '';
		}

		return join("/", $parts) . '/';
	}
}
